const BaseError = require('./validation/errors/base-error');
const {isVerboseMode} = require('./utils');
const {red, yellow, green, bgRed} = require('colorette');

const emptyLine = length => new Array(length).fill(' ').join('');
const multiline = (color, msg) => {
    msg = ` ${msg}`;
    msg = msg + emptyLine(120 - msg.length);

    return `\n${color(emptyLine(msg.length))}\n${color(msg)}\n${color(emptyLine(msg.length))}`;
};

class ErrorHandler {
    constructor() {
        this.failedThemes = [];

        process.on('beforeExit', code => {
            this.onProcess();
        });
    }

    /**
     * Show a message about failed themes
     */
    onProcess() {
        if (this.failedThemes.length) {
            let msg = '[ERROR] Assets build';

            if (this.failedThemes.length === 1) {
                msg = `${msg} for "${this.failedThemes[0]}" theme failed.`;
            } else {
                msg = `${msg} for "${this.failedThemes.join(', ')}" themes failed.`;
            }
            console.error(multiline(bgRed, msg));
            process.exit(1);
        }
    }

    /**
     * Executes when the Webpack Compiler will complete a compilation
     * In case of empty Webpack Config this method will not run
     *
     * @param {Object} stats
     */
    onBuildComplete(stats) {
        const failed = stats?.compilation?.errors.length !== 0;

        if (failed) {
            console.error(multiline(bgRed, '[ERROR] Assets build failed.'));
            process.exit(1);
        }
    }

    /**
     * @param {Array|string} errors
     * @param {string} [failedTheme]
     */
    displayError(errors, failedTheme) {
        if (typeof failedTheme === 'string') {
            this.failedThemes.push(failedTheme);
        }

        if (typeof errors === 'string') {
            errors = [errors];
        }

        const errorMessage = errors.map(err => {
            const highlightPrefix = str => {
                const prefix = str.substring(0, str.search(':'));
                return red(prefix) + str.substring(str.search(':'));
            };

            if (err instanceof BaseError) {
                err.message = `${err.message}\n${green(err.extra)}`;
            }

            let msg = err.message;

            if (isVerboseMode()) {
                msg = highlightPrefix(err.stack);
            } else {
                const command = yellow('php bin/console oro:assets:build --env=dev --verbose');
                const info = `${green(`Run the command ${command} to see more information`)}`;

                msg = highlightPrefix(`${err.name}: ${msg}\n${info}`);
            }

            return `${msg}\n`;
        });

        console.error(errorMessage.join('\n'));
    }
}

module.exports = ErrorHandler;
