const path = require('path');
const merge = require('deepmerge');
const ModulesConfigLoader = require('./modules-config-loader');

class LayoutModulesConfigLoader extends ModulesConfigLoader {
    /**
     * @inheritdoc
     */
    loadConfig(theme, filePath) {
        let themeConfig = super.loadConfig(theme, path.join('Resources/views/layouts/', theme, filePath));
        themeConfig = merge(themeConfig, super.loadConfig(theme, path.join('templates/layouts/', theme, filePath)));
        // recursive process parent theme
        const {parent: parentTheme} = this.themes[theme];
        if (typeof parentTheme === 'string') {
            const processedFiles = this.processedFiles;
            const parentThemeConfig = this.loadConfig(parentTheme, filePath);

            themeConfig = merge(parentThemeConfig, themeConfig);
            // processedFiles from parent theme is added to processedFiles of current theme
            this._processedFiles = [...processedFiles, ...this.processedFiles];
        }
        return themeConfig;
    }

    /**
     * All build names:
     *  - based on theme names
     *  - and theme's extra js builds
     *
     * @return {string[]}
     */
    get buildNames() {
        const buildNames = super.buildNames;
        this.themeNames.forEach(theme => {
            buildNames.push(...this.extraJSBuildNames(theme));
        });
        return buildNames;
    }

    /**
     * Collect extra js-build names for a theme
     *
     * @param {string} theme name on the theme
     * @return {string[]}
     */
    extraJSBuildNames(theme) {
        const {extra_js_builds: extraJSBuilds = []} = this._themes[theme] || {};
        return [...extraJSBuilds.map(suffix => `${theme}-${suffix}`)];
    }

    /**
     * Check if buildName is an extra js build of layout theme
     *
     * @param {string} buildName name of the build
     * @return {boolean}
     */
    isExtraJSBuild(buildName) {
        const [, suffix] = this.splitBuildName(buildName);
        return suffix !== void 0;
    }

    /**
     * Splits build name on parts theme name and suffix (name of extra js build)
     * @param {string} buildName
     * @return {[string]|[string, string]}
     */
    splitBuildName(buildName) {
        // suffix can not contain '-'
        const marches = buildName.match(/(.+)-([^\-]+)?/);
        const result = [];
        if (marches) {
            const [, theme, suffix] = marches;
            const {extra_js_builds: extraJSBuilds = []} = this._themes[theme] || {};
            if (extraJSBuilds.includes(suffix)) {
                result.push(marches[1], marches[2]);
            } else {
                result.push(buildName);
            }
        } else {
            result.push(buildName);
        }
        return result;
    }
}

module.exports = LayoutModulesConfigLoader;
