const path = require('path');
const fs = require('fs');
const merge = require('deepmerge');
const yaml = require('js-yaml');
const validation = require('../validation');

// merge only unique items
const arrayMerge = (target, source) => target.concat(source.filter(item => !target.includes(item)));

class ModulesConfigLoader {
    /**
     * @returns {Array}
     */
    get themes() {
        return this._themes;
    }

    /**
     * @returns {Array}
     */
    get themeNames() {
        return Object.keys(this._themes);
    }

    /**
     * @returns {Array}
     */
    get buildNames() {
        return this.themeNames;
    }

    /**
     * @returns {Array}
     */
    get processedFiles() {
        return [...this._processedFiles];
    }

    /**
     * @param {Array} bundles Array of ordered symfony bundle paths
     * @param {string|Array} themesLocation Path inside the bundle, where to find the theme
     * @param {string} themeInfoFileName Yaml File name with theme info
     */
    constructor(bundles, themesLocation, themeInfoFileName) {
        this._bundles = bundles;
        if (!Array.isArray(themesLocation)) {
            themesLocation = [themesLocation];
        }

        const themes = {};
        const self = this;
        themesLocation.forEach(themesLocation => {
            self._collectThemes(themes, themesLocation, themeInfoFileName);
        });

        this._themes = themes;
        this._processedFiles = [];
    }

    /**
     * Collects list of themes with their parents into given storage(themes param)
     *
     * @param {Object} themes
     * @param {string} themesLocation
     * @param {string} themeInfoFileName
     * @private
     */
    _collectThemes(themes, themesLocation, themeInfoFileName) {
        this._bundles.forEach(bundle => {
            const source = bundle + themesLocation;

            if (!fs.existsSync(source)) return;

            fs.readdirSync(source).forEach(name => {
                const themePath = path.resolve(source, name);
                if (!fs.lstatSync(themePath).isDirectory()) {
                    return;
                }
                const themeFile = path.resolve(themePath, themeInfoFileName);
                if (!fs.existsSync(themeFile)) return;

                const theme = yaml.load(fs.readFileSync(themeFile, 'utf8'));
                themes[name] = merge(themes[name] || {}, theme, {arrayMerge});
            });
        });
    }

    /**
     * @param {string} theme Theme name
     * @param {string|string[]} filePath Path (or paths with fallback) to the file inside bundle directory where to find the configs
     * @return {Object} Merged Configs loaded from all the bundles Yaml files matched by filePath
     */
    loadConfig(theme, filePath) {
        let configs = {};
        const filePaths = [].concat(filePath);

        this._processedFiles = [];
        this._bundles.forEach(bundle => {
            let absolutePath;

            for (const file of filePaths) {
                absolutePath = path.resolve(bundle, file);
                if (fs.existsSync(absolutePath)) {
                    break;
                }
                absolutePath = void 0;
            }

            if (absolutePath) {
                const rawDoc = fs.readFileSync(absolutePath, 'utf8');
                const doc = yaml.load(rawDoc);

                this._processedFiles.push(absolutePath);

                validation.checkSchema(absolutePath, doc, theme);
                configs = merge(configs, doc);
            }
        });

        return configs;
    }
}

module.exports = ModulesConfigLoader;
