const webpack = require('webpack');
const printf = require('printf');
const AppConfigLoader = require('./app-config-loader');
const AppModulesFileWriter = require('./writer/app-modules-file-writer');
const CleanupStatsPlugin = require('./plugin/stats/cleanup-stats-plugin');
const AfterWebpackLogsPlugin = require('./plugin/logs/after-webpack-logs-plugin');
const ConfigsFileWriter = require('./writer/configs-file-writer');
const EntryPointFileWriter = require('./writer/scss-entry-point-file-writer');
const LayoutModulesConfigLoader = require('./modules-config/layout-modules-config-loader');
const LayoutStyleLoader = require('./style/layout-style-loader');
const MapModulesPlugin = require('./plugin/map/map-modules-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const ModulesConfigLoader = require('./modules-config/modules-config-loader');
const DynamicImportsFileWriter = require('./writer/dynamic-imports-file-writer');
const CssMinimizerPlugin = require('css-minimizer-webpack-plugin');
const prepareModulesShim = require('./prepare-modules-shim');
const AdminStyleLoader = require('./style/admin-style-loader');
const ThemeConfigFactory = require('./theme-config-factory');
const path = require('path');
const fs = require('fs');
const prepareModulesMap = require('./plugin/map/prepare-modules-map');
const resolve = require('enhanced-resolve');
const {merge: webpackMerge} = require('webpack-merge');
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin;
const RtlCssWebpackPlugin = require('rtlcss-webpack-plugin');
const validation = require('./validation');
const EventEmitter = require('events');
const ErrorHandler = require('./error-handler');

require('resolve-url-loader');

class ConfigBuilder {
    constructor() {
        this._projectPath = '';
        this._enableLayoutThemes = false;
        this._defaultLayoutThemes = null;
        this.emitter = new EventEmitter();
        this.errorHandler = new ErrorHandler();
        this.addListeners();
        this.setPublicPath('public/');
        this.setAdminTheme('admin.oro');
        this.setVersionFormat('%s?v=%s');
        this.setBabelConfig({
            sourceType: 'unambiguous',
            presets: [
                [
                    '@babel/preset-env', {
                        useBuiltIns: 'usage',
                        corejs: {
                            version: 3,
                            proposals: true
                        }
                    }
                ]
            ],
            plugins: [
                '@babel/plugin-transform-runtime'
            ]
        });
    }

    addListeners() {
        this.emitter.on('publicPath:updated', path => validation.setPublicPath(path));
        this.emitter.on('build:complete', stats => this.errorHandler.onBuildComplete(stats));
    }

    /**
     * Set Symfony public directory path related to application root folder
     * @param {string} publicPath
     * @returns {ConfigBuilder}
     */
    setPublicPath(publicPath) {
        this._publicPath = publicPath;
        this.emitter.emit('publicPath:updated', publicPath);
        return this;
    }

    /**
     * Set Symfony cache directory path related to application root folder
     * @param {string} cachePath
     * @returns {ConfigBuilder}
     */
    setCachePath(cachePath) {
        this._cachePath = cachePath;
        return this;
    }

    /**
     * Specifies a sprintf pattern that will be used with the version option to construct an asset’s path.
     * By default ts `%s?v=%s`
     * @param {string} versionFormat
     * @returns {ConfigBuilder}
     */
    setVersionFormat(versionFormat) {
        this._versionFormat = versionFormat;
        this.emitter.emit('versionFormat:updated', versionFormat);
        return this;
    }

    /**
     * Set active admin (management console) theme. Out of the box there are 2 themes ("admin.oro" and "admin.demo")
     *
     * @param {string} adminTheme
     * @returns {ConfigBuilder}
     */
    setAdminTheme(adminTheme) {
        const themeNameParts = adminTheme.split('.');
        if (themeNameParts[0] !== 'admin' || themeNameParts.length !== 2) {
            throw new Error('Admin theme name should be in a format "admin.{themeName}", for example "admin.oro"');
        }

        this._adminTheme = adminTheme;
        this.emitter.emit('adminTheme:updated', adminTheme);
        return this;
    }

    /**
     * Enable build of Layout themes.
     */
    enableLayoutThemes(themes) {
        if (themes) {
            this._defaultLayoutThemes = themes;
            this.emitter.emit('defaultLayoutThemes:updated', themes);
        }
        this._enableLayoutThemes = true;
        this.emitter.emit('enableLayoutThemes:updated', true);
        return this;
    }

    /**
     * Set babel config
     * @param {object} babelConfig
     * @returns {ConfigBuilder}
     */
    setBabelConfig(babelConfig) {
        this._babelConfig = babelConfig;
        this.emitter.emit('babelConfig:updated', babelConfig);
        return this;
    }

    /**
     * Returns callback that loads webpack configs based on environment and command arguments variables
     * @returns {Function}
     */
    getWebpackConfig() {
        return (env = {}, args = {}) => {
            this._initialize(args, env);

            let commonConfig = {};
            try {
                commonConfig = this._getCommonWebpackConfig(args, env);
            } catch (e) {
                this.errorHandler.displayError(e);
                process.exit(1);
            }

            const webpackConfigs = [];
            const requestedBuildNames = env.theme ? env.theme.split(',') : [];
            const buildNames = this._getBuildNames(requestedBuildNames);

            const buildErrors = [];
            validation.jsmodulesValidator.emitter.on('error', error => buildErrors.push(error));
            validation.assetsValidation.emitter.on('error', error => buildErrors.push(error));
            buildNames.forEach(buildName => {
                let buildConfig;
                // flush all collected errors from previews builds
                buildErrors.splice(0, buildErrors.length);

                try {
                    buildConfig = this._getThemeWebpackConfig(buildName, args, env);
                } catch (error) {
                    buildErrors.push(error);
                }

                if (buildErrors.length) {
                    this.errorHandler.displayError(buildErrors, buildName);
                    return;
                }

                if (buildConfig) {
                    webpackConfigs.push(webpackMerge(buildConfig, commonConfig));
                }
            });

            return webpackConfigs;
        };
    }

    get resolvedProjectPath() {
        if (this._resolvedProjectPath == undefined) {
            this._resolvedProjectPath = path.resolve(this._projectPath);
        }
        return this._resolvedProjectPath;
    }

    get resolvedPublicPath() {
        if (this._resolvedPublicPath === undefined) {
            this._resolvedPublicPath = path.resolve(this._publicPath);
        }
        return this._resolvedPublicPath;
    }

    get resolvedNodeModulesPath() {
        if (this._resolvedNodeModulesPath === undefined) {
            this._resolvedNodeModulesPath = path.resolve('node_modules');
        }
        return this._resolvedNodeModulesPath;
    }

    get assetVersion() {
        if (this._assetVersion === undefined) {
            const filePath = path.join(this.resolvedPublicPath, '/build/build_version.txt');
            this._assetVersion = fs.existsSync(filePath) ? String(fs.readFileSync(filePath)) : null;
        }
        return this._assetVersion;
    }

    _getBuildNames(requestedBuildNames = []) {
        const buildNames = [];
        requestedBuildNames.forEach(buildName => this._validateBuildName(buildName));

        // Admin themes
        if (!requestedBuildNames.length) {
            buildNames.push(this._adminTheme);
        } else {
            requestedBuildNames.forEach(buildName => {
                if (this._isAdminTheme(buildName)) {
                    buildNames.push(buildName);
                }
            });
        }

        // Layout Themes
        if (this._enableLayoutThemes) {
            if (!requestedBuildNames.length) {
                // build all layout themes
                const themes = this._defaultLayoutThemes || this._appConfig.themes;
                buildNames.push(...themes);
                themes.forEach(theme => {
                    buildNames.push(...this._layoutModulesConfigLoader.extraJSBuildNames(theme));
                });
            } else {
                // build single layout theme
                requestedBuildNames.forEach(buildName => {
                    if (this._layoutModulesConfigLoader.buildNames.includes(buildName)) {
                        buildNames.push(buildName);
                    }
                });
            }
        }

        return buildNames;
    }

    _getCommonWebpackConfig(args, env) {
        const stats = env.stats || {
            hash: false,
            version: false,
            // Do not write the information about files emitted from node_modules or public bundles,
            // and hide css/*/*.js files from the output.
            excludeAssets: [/^bundles\//, /^\.\.\/_static\//, /^css\/.+\.js/],
            children: false,
            entrypoints: false,
            performance: this._isProduction,
            chunks: false,
            modules: false,
            source: false,
            publicPath: true,
            builtAt: false,
            warnings: false
        };

        const webpackConfig = {
            watchOptions: {
                aggregateTimeout: 200,
                ignored: /[\/\\]node_modules[\/\\].*\.js$/
            },
            stats,
            output: {
                filename: '[name].js',
                // Because we use third party libraries 'chunkFilename' should include only [name]
                chunkFilename: this._getVersionedPath('chunk/[name].js', this.assetVersion)
            },
            devtool: !env.skipSourcemap && 'inline-cheap-module-source-map',
            mode: 'development',
            optimization: {
                moduleIds: 'named',
                splitChunks: {
                    cacheGroups: {
                        defaultVendors: false
                    }
                }
            },
            resolveLoader: {
                modules: [
                    this.resolvedPublicPath,
                    path.join(__dirname, './loader'),
                    path.join(this.resolvedPublicPath, '/bundles'),
                    this.resolvedNodeModulesPath
                ]
            },
            module: {
                noParse: [
                    /[\/\\]bundles[\/\\]\.*[\/\\]lib[\/\\](?!chaplin|bootstrap|jquery\.dialog).*\.js$/
                ],
                rules: [
                    {
                        test: /\.s?css$/,
                        use: [{
                            loader: args.hot ? 'style-loader' : MiniCssExtractPlugin.loader
                        }, {
                            loader: 'css-loader',
                            options: {
                                importLoaders: 1,
                                sourceMap: true,
                                // can't use esModule since resolve-url-loader needs file path to start with '~'
                                esModule: false
                            }
                        }, {
                            loader: 'resolve-url-loader'
                        }, {
                            loader: 'postcss-loader',
                            options: {
                                sourceMap: true,
                                postcssOptions: {
                                    plugins: [
                                        require('autoprefixer')
                                    ]
                                }
                            }
                        }, {
                            loader: 'sass-loader',
                            options: {
                                implementation: require("sass"),
                                sassOptions: {
                                    includePaths: [
                                        path.join(this.resolvedPublicPath, '/bundles')
                                    ],
                                    outputStyle: 'expanded'
                                },
                                sourceMap: true
                            }
                        }]
                    },
                    {
                        test: /\.(eot|ttf|woff|woff2|cur|ico|svg|png|jpg|gif)$/,
                        loader: 'url-loader',
                        options: {
                            limit: 1,
                            emitFile: true,
                            outputPath: '../_static/',
                            publicPath: '../../_static/',
                            name: this._getVersionedPath('[path][name].[ext]', this.assetVersion)
                        }
                    }
                ]
            },
            performance: {hints: false},
            plugins: [
                new MiniCssExtractPlugin({
                    filename: '[name].css'
                }),
                new CleanupStatsPlugin(),
                // Ignore all locale files of moment.js
                new webpack.IgnorePlugin({
                    resourceRegExp: /^\.[\/\\]locale$/,
                    contextRegExp: /moment$/
                }),
                new webpack.optimize.MinChunkSizePlugin({
                    minChunkSize: 30000 // Minimum number of characters
                }),
                new AfterWebpackLogsPlugin(
                    stats => this.emitter.emit('build:complete', stats)
                )
            ]
        };

        if (env.analyze) {
            webpackConfig.plugins.push(new BundleAnalyzerPlugin());
        }

        if (!env.skipJS && env.withBabel) {
            webpackConfig.module.rules.push({
                test: /\.js$/,
                exclude: [
                    /[\/\\]platform[\/\\]build[\/\\]/,
                    /[\/\\]node_modules[\/\\]/,
                    /[\/\\]bundles[\/\\].+[\/\\]lib[\/\\]?/
                ],
                use: {
                    loader: 'babel-loader',
                    options: this._babelConfig
                }
            });
        }

        if (args.hot) {
            const https = this._appConfig.devServerOptions.https;
            const schema = https ? 'https' : 'http';
            const devServerHost = this._appConfig.devServerOptions.host;
            const devServerPort = this._appConfig.devServerOptions.port;
            webpackConfig.devServer = {
                contentBase: this.resolvedPublicPath,
                host: devServerHost,
                port: devServerPort,
                https: https,
                compress: true,
                stats: webpackConfig.stats,
                disableHostCheck: true,
                clientLogLevel: 'error',
                headers: {
                    'Access-Control-Allow-Origin': '*'
                }
            };
            webpackConfig.output.publicPath = `${schema}://${devServerHost}:${devServerPort}/`;
        }

        // Additional setting for production mode
        if (this._isProduction) {
            webpackConfig.devtool = false;
            webpackConfig.plugins.push(new CssMinimizerPlugin());
        }

        return webpackConfig;
    }

    _getThemeWebpackConfig(buildName, args, env) {
        let {skipCSS, skipJS, skipRTL} = env;
        let themeDefinition;
        let jsBuildConfig;
        let buildPublicPath;
        if (this._isAdminTheme(buildName)) {
            themeDefinition = this._modulesConfigLoader.themes[buildName.split('.')[1]];
            buildPublicPath = '/build/admin/';
            const jsModulesConfig = this._themeConfigFactory.loadConfig(
                buildName,
                [
                    'Resources/config/oro/jsmodules.yml',
                    'Resources/config/jsmodules.yml',
                    'config/oro/jsmodules.yml'
                ]
            );            validation.jsmodulesValidator.checkFullSchema(
                jsModulesConfig,
                this._themeConfigFactory?._configLoader.processedFiles,
                buildName
            );
            jsBuildConfig = this._themeConfigFactory.create(buildPublicPath, jsModulesConfig);
        } else if (this._layoutModulesConfigLoader.isExtraJSBuild(buildName)) {
            const [theme, suffix] = this._layoutModulesConfigLoader.splitBuildName(buildName);
            skipCSS = true;
            themeDefinition = this._layoutModulesConfigLoader.themes[theme];
            buildPublicPath = `/build/${buildName}/`;
            const baseConfig = this._layoutThemeConfigFactory.loadConfig(theme, 'config/jsmodules.yml');
            const extraConfig = this._layoutThemeConfigFactory.loadConfig(theme, `config/jsmodules-${suffix}.yml`);
            const jsModulesConfig = this._layoutThemeConfigFactory.extendConfig(baseConfig, extraConfig);
            validation.jsmodulesValidator.checkFullSchema(
                jsModulesConfig,
                this._layoutThemeConfigFactory?._configLoader.processedFiles,
                buildName
            );
            jsBuildConfig = this._layoutThemeConfigFactory.create(buildPublicPath, jsModulesConfig);
        } else {
            themeDefinition = this._layoutModulesConfigLoader.themes[buildName];
            buildPublicPath = `/build/${buildName}/`;
            const jsModulesConfig = this._layoutThemeConfigFactory.loadConfig(buildName, 'config/jsmodules.yml');
            validation.jsmodulesValidator.checkFullSchema(
                jsModulesConfig,
                this._layoutThemeConfigFactory?._configLoader.processedFiles,
                buildName
            );
            jsBuildConfig = this._layoutThemeConfigFactory.create(buildPublicPath, jsModulesConfig);
        }
        const {rtl_support: rtlSupport = false} = themeDefinition;
        const resolvedBuildPath = path.join(this.resolvedPublicPath, buildPublicPath);

        const resolverConfig = {
            modules: [
                resolvedBuildPath,
                this.resolvedPublicPath,
                path.join(this.resolvedProjectPath, '/assets'),
                path.join(this.resolvedPublicPath, '/bundles'),
                path.join(this.resolvedPublicPath, '/js'),
                this.resolvedNodeModulesPath
            ],
            alias: jsBuildConfig.aliases,
            cacheWithContext: true,
            symlinks: false
        };
        const resolver = (resolver => {
            return moduleName => resolver({}, '', moduleName, {});
        })(resolve.create.sync({...resolverConfig}));

        const plugins = [];
        if (rtlSupport && !skipCSS && !skipRTL) {
            plugins.push(new RtlCssWebpackPlugin({
                filename: '[name].rtl.css'
            }));
        }

        const cssEntryPoints = !skipCSS ? this._getCssEntryPoints(buildName, buildPublicPath) : {};
        const jsEntryPoints = !skipJS ? jsBuildConfig.entry : {};

        const entryPoints = {...cssEntryPoints, ...jsEntryPoints};
        if (Object.keys(entryPoints).length === 0) {
            return;
        }

        return {
            entry: entryPoints,
            name: buildName + ' theme',
            output: {
                publicPath: buildPublicPath,
                path: resolvedBuildPath
            },
            context: this.resolvedPublicPath,
            resolve: {
                ...resolverConfig,
                plugins: [
                    new MapModulesPlugin(prepareModulesMap(resolver, jsBuildConfig.map))
                ]
            },
            plugins,
            module: {
                rules: [
                    {
                        test: /[\/\\]configs\.json$/,
                        loader: 'config-loader',
                        options: {
                            resolver,
                            relativeTo: this.resolvedPublicPath
                        }
                    },
                    ...prepareModulesShim(resolver, jsBuildConfig.shim)
                ]
            }
        };
    }

    _initialize(args, env) {
        const entryPointFileWriter = new EntryPointFileWriter(this._publicPath);

        this._isProduction = args.mode === 'production';
        this._symfonyEnv = env.symfony;
        this._appConfig = AppConfigLoader.getConfig(this._cachePath, this._symfonyEnv);
        this._appConfig.paths.push(this.resolvedProjectPath);

        this._modulesConfigLoader = new ModulesConfigLoader(
            this._appConfig.paths,
            ['/Resources/public/themes/', '/public/themes/admin/'],
            'settings.yml'
        );
        this._adminThemes = this._modulesConfigLoader.themeNames.map(themeName => 'admin.' + themeName);
        this._adminStyleLoader = new AdminStyleLoader(this._modulesConfigLoader, entryPointFileWriter);
        this._themeConfigFactory = new ThemeConfigFactory(
            this._modulesConfigLoader,
            new DynamicImportsFileWriter(this._publicPath),
            new AppModulesFileWriter(this._publicPath),
            new ConfigsFileWriter(this._publicPath)
        );

        this._layoutModulesConfigLoader = new LayoutModulesConfigLoader(
            this._appConfig.paths,
            ['/Resources/views/layouts/', '/templates/layouts/'],
            'theme.yml'
        );
        this._layoutStyleLoader = new LayoutStyleLoader(this._layoutModulesConfigLoader, entryPointFileWriter);
        this._layoutThemeConfigFactory = new ThemeConfigFactory(
            this._layoutModulesConfigLoader,
            new DynamicImportsFileWriter(this._publicPath),
            new AppModulesFileWriter(this._publicPath),
            new ConfigsFileWriter(this._publicPath)
        );
    }

    _getVersionedPath(name, assetVersion) {
        if (!assetVersion) {
            return name;
        }
        return printf(this._versionFormat, name, assetVersion);
    }

    _getCssEntryPoints(theme, buildPath) {
        if (this._isAdminTheme(theme)) {
            return this._adminStyleLoader.getThemeEntryPoints(theme.split('.')[1], buildPath);
        }

        return this._layoutStyleLoader.getThemeEntryPoints(theme, buildPath);
    }

    _validateBuildName(buildName) {
        const buildNames = [...this._adminThemes];
        if (this._enableLayoutThemes) {
            buildNames.push(...this._layoutModulesConfigLoader.buildNames);
        }
        if (buildName === 'admin') {
            throw new Error(
                'The "admin" is a reserved word and cannot be used as a theme name.'
            );
        }
        if (buildName !== undefined && !buildNames.includes(buildName)) {
            throw new Error(
                'Theme "' + buildName + '" doesn\'t exists. Existing themes:' + buildNames.join(', ')
            );
        }
    }

    _isAdminTheme(theme) {
        return this._adminThemes.includes(theme);
    }
}

module.exports = new ConfigBuilder;
