const path = require('path');
const wildcard = require('wildcard');
const _ = require('underscore');
const {isProdMode} = require('../utils');
const {assetsValidation} = require('../validation');
const StylesError = require('../validation/errors/styles-error');

/**
 * @typedef ThemeGroupConfig
 * @type {Object}
 * @property {Array.<string|Object>} inputs - list of input resources for import into root SCSS file
 * @property {Array.<string>} entries - list of direct entry resources
 * @property {string} output - name or out put resource
 * @property {[boolean]} auto_rtl_inputs - whether theme support RTL
 */

class StyleLoader {
    /**
     * @param {YamlConfigLoader} configLoader
     * @param {SCSSEntryPointFileWriter} entryPointFileWriter
     */
    constructor(configLoader, entryPointFileWriter) {
        this._configLoader = configLoader;
        this._entryPointFileWriter = entryPointFileWriter;
    }

    /**
     * @param {string} theme Theme name
     * @param {string} buildPath Build path
     * @return {Object} List of Webpack entry-points
     */
    getThemeEntryPoints(theme, buildPath) {
        const {themeConfig, settings = {}} = this._fetchThemeConfig(theme);

        assetsValidation.checkFullSchema(themeConfig, this._configLoader.processedFiles, theme);

        const entryPoints = {};
        const writingOptions = {};
        for (const [group, config] of Object.entries(themeConfig)) {
            let {inputs, entries = [], output, auto_rtl_inputs: rtlMasks = []} = config;

            if (isProdMode() && output === void 0) {
                throw new StylesError('output', group, theme);
            }

            if (isProdMode() && inputs === void 0) {
                throw new StylesError('inputs', group, theme);
            }

            inputs = this._overrideInputs(inputs);
            inputs = this._sortInputs(inputs);
            inputs = this._applyInputsBasePathPrefix(inputs);

            if (settings.rtlSupport) {
                writingOptions.ignoreRTLInputs = _.difference(inputs, this._matchInputs(rtlMasks, inputs));
            }

            const entryPointName = output.replace(/\.[^/.]+$/, '');
            const filePath = path.join(buildPath, output);
            entryPoints[entryPointName] = [
                ...entries,
                this._entryPointFileWriter.write('./../../', inputs, filePath, writingOptions)
            ];
        }
        return entryPoints;
    }

    /**
     * Extracts theme configuration with its settings
     *
     * @param {string} themeName
     * @return {{themeConfig: Object.<string, ThemeGroupConfig>, settings: {rtlSupport: boolean}}}
     * @abstract
     * @protected
     */
    _fetchThemeConfig(themeName) {
        throw new Error('Method `_fetchThemeConfig` has to be implemented in extends');
    }

    /**
     * Override inputs
     *
     * @param {(string|Object)[]} inputs
     * @returns {string[]} List of inputs
     * @protected
     */
    _overrideInputs(inputs) {
        const mappedInputs = [];
        inputs.forEach(input => {
            if (typeof input === 'string') {
                mappedInputs.push(input);
            } else {
                const [oldInput, newInput] = Object.entries(input)[0];
                const oldInputIndex = mappedInputs.findIndex(item => item === oldInput);

                if (oldInputIndex === -1) {
                    // old input does not exist anymore
                    mappedInputs.push(newInput);
                } else if (newInput) {
                    // replace input
                    mappedInputs[oldInputIndex] = newInput;
                } else {
                    // delete input
                    mappedInputs.splice(oldInputIndex, 1);
                }
            }
        });
        return mappedInputs;
    }

    /**
     * Sort inputs, so first will be settings, then variables, then other files
     *
     * @param {string[]} inputs
     * @returns {string[]} List of ordered inputs
     * @protected
     */
    _sortInputs(inputs) {
        const primarySettingsInputs = [];
        const settingsInputs = [];
        const primaryVariablesInputs = [];
        const variablesInputs = [];
        const restInputs = [];

        inputs.forEach(input => {
            if (input.indexOf('/settings/primary-settings') > 0) {
                primarySettingsInputs.push(input);
            } else if (input.indexOf('/settings/') > 0) {
                settingsInputs.push(input);
            } else if (input.indexOf('/variables/primary-variables') > 0) {
                primaryVariablesInputs.push(input);
            } else if (input.indexOf('/variables/') > 0) {
                variablesInputs.push(input);
            } else {
                restInputs.push(input);
            }
        });

        return [
            ...primarySettingsInputs,
            ...settingsInputs,
            ...primaryVariablesInputs,
            ...variablesInputs,
            ...restInputs
        ];
    }

    /**
     * Filter inputs that matches any mask from the list
     *
     * @param {string[]} masks list of wildcard masks
     * @param {string[]} inputs
     * @return {string[]} matched inputs
     * @protected
     */
    _matchInputs(masks, inputs) {
        masks = this._applyInputsBasePathPrefix(masks);
        masks = masks.map(mask => wildcard(mask));

        const whiteListedInputs = masks.reduce((include, mask) => {
            include.push(...mask.match(inputs));
            return include;
        }, []);

        return _.unique(whiteListedInputs);
    }

    /**
     * Considering base path as application's root.
     * Bundles based *.scss sources go with '../' prefix.
     *
     * @param {string[]} inputs
     * @private
     */
    _applyInputsBasePathPrefix(inputs) {
        const processedInputs = [];

        inputs.forEach(input => {
            if (input.indexOf('bundles') === 0) {
                input = '../' + input;
            }
            processedInputs.push(input);
        });

        return processedInputs;
    }
}

module.exports = StyleLoader;
