const JSModulesExtraModulesError = require('./validation/errors/jsmodules-extra-modules-error');

class ThemeConfigFactory {
    /**
     * @param {ModulesConfigLoader} configLoader
     * @param {DynamicImportsFileWriter} dynamicImportsFileWriter
     * @param {AppModulesFileWriter} appModulesFileWriter
     * @param {ConfigsFileWriter} configsFileWriter
     */
    constructor(configLoader, dynamicImportsFileWriter, appModulesFileWriter, configsFileWriter) {
        this._configLoader = configLoader;
        this._dynamicImportsFileWriter = dynamicImportsFileWriter;
        this._appModulesFileWriter = appModulesFileWriter;
        this._configsFileWriter = configsFileWriter;
    }
    /**
     * @param {string} theme Theme name
     * @param {string|string[]} configFilepath Path (or paths with fallback) to yaml config file in a bundle
     * @return {Object} Merged Configs loaded from all the bundles Yaml files matched by filePath
     */
    loadConfig(theme, configFilepath) {
        return this._configLoader.loadConfig(theme, configFilepath);
    }

    extendConfig(baseConfig, extraConfig) {
        const {
            aliases = {},
            configs,
            map = {},
            shim = {}
        } = baseConfig;

        const {
            ['app-modules']: appModules,
            ['dynamic-imports']: dynamicImports,
            entry,
            ...rest
        } = extraConfig;

        const beyondKeys = Object.keys(rest);
        if (beyondKeys.length) {
            throw new JSModulesExtraModulesError(beyondKeys);
        }

        return {
            aliases,
            'app-modules': appModules,
            configs,
            'dynamic-imports': dynamicImports,
            entry,
            map,
            shim
        };
    }

    /**
     * @param {string} buildPath Path to theme build folder
     * @param {Object} jsModulesConfig configuration loaded from jsmodules files and merged together
     * @return {Object} webpack configuration fragment
     */
    create(buildPath, jsModulesConfig) {
        const {
            entry,
            map = {},
            shim = {},
            configs,
            aliases = {},
            ['app-modules']: appModules = [],
            ['dynamic-imports']: dynamicImports
        } = jsModulesConfig;

        if (configs) {
            this._configsFileWriter.write(jsModulesConfig.configs, buildPath);
        }

        if (appModules) {
            this._appModulesFileWriter.write(appModules, buildPath);
        }

        if (dynamicImports) {
            this._dynamicImportsFileWriter.write(dynamicImports, buildPath);
        }

        return {entry, map, shim, aliases};
    }
}

module.exports = ThemeConfigFactory;
