const {isProdMode} = require('../utils');
const fs = require('fs');
const path = require('path');
const schema = require('./schemas/assets-schema');
const fullSchema = require('./schemas/assets-schema-full');
const schemaValidator = require('./schema-validator');
const EventEmitter = require('events');
const emitter = new EventEmitter();
const AssetsSchemaError = require('./errors/assets-schema-error');
const AssetsInputFileError = require('./errors/assets-input-file-error');

module.exports = Object.assign({}, schemaValidator, {
    emitter,

    /**
     * @param {string} filePath
     * @param {Object} doc
     * @param {string} theme
     * @returns {boolean|undefined}
     */
    checkSchema(filePath, doc, theme) {
        if (isProdMode()) {
            return;
        }
        const result = this.validateSchema(schema, doc);

        if (!result.valid) {
            const error = new AssetsSchemaError(result.formattedError, [filePath], theme);

            this.emitter.emit('error', error);
        }

        return result.valid;
    },

    /**
     * @param {Object} doc
     * @param {Array} files
     * @param {string} theme
     * @returns {boolean|undefined}
     */
    checkFullSchema(doc, files = [], theme) {
        if (isProdMode()) {
            return;
        }
        const result = this.validateSchema(fullSchema, doc);

        if (!result.valid) {
            const error = new AssetsSchemaError(result.formattedError, files, theme);

            this.emitter.emit('error', error);
        }

        return result.valid;
    },

    /**
     * @param {string} filePath
     * @param {Object} doc
     * @param {string} theme
     * @returns {boolean|undefined}
     */
    checkInputsExist(filePath, doc, theme) {
        if (isProdMode()) {
            return;
        }

        let valid = true;
        for (const [, props] of Object.entries(doc)) {
            if (!Array.isArray(props.inputs)) {
                continue;
            }

            const files = [];
            props.inputs.forEach(input => {
                if (typeof input === 'object') {
                    const newPath = Object.values(input)[0];

                    if (newPath === '~') {
                        return;
                    }
                    input = newPath;
                }

                const fullBasePath = path.resolve(input);
                const fullPath = path.resolve(this._publicPath + input);
                // skip the path to global node modules,
                // e.g. '~bootstrap/scss/bootstrap'
                if (!input.startsWith('~') && !fs.existsSync(fullPath) && !fs.existsSync(fullBasePath)) {
                    files.push(input);
                }
            });
            if (files.length) {
                valid = false;
                files.forEach(notExistFile => {
                    const error = new AssetsInputFileError(notExistFile, filePath);

                    this.emitter.emit('error', error);
                });
            }
        }
        return valid;
    }
});
