const path = require('path');
const fs = require('fs');

class SCSSEntryPointFileWriter {
    /**
     * @param {string} publicPath
     */
    constructor(publicPath) {
        this._publicPath = publicPath;
    }

    /**
     * Write entry point file and return file path
     *
     * @param {string} baseInputPath base path for input files
     * @param {Array} inputs List of inputs
     * @param {string} output Output file path
     * @param {Object} options writing options
     * @returns {string} JS file path of an output file
     */
    write(baseInputPath, inputs, output, {ignoreRTLInputs = []}) {
        let content = '';
        inputs.forEach(input => {
            const ignoreRTL = ignoreRTLInputs.indexOf(input) !== -1;
            input = input.replace(/\.[^/.]+$/, '');
            // don't add the base path to global node modules,
            // e.g. '~bootstrap/scss/bootstrap'
            const basePath = input.startsWith('~') ? '' : baseInputPath;
            let importModule = `@import "${basePath}${input}";\n`;
            if (ignoreRTL) {
                importModule = `/*rtl:begin:ignore*/\n${importModule}/*rtl:end:ignore*/\n`;
            }
            content += importModule;
        });
        const scssFilepath = path.resolve(this._publicPath + output + '.scss');
        fs.mkdirSync(path.dirname(scssFilepath), {recursive: true});
        fs.writeFileSync(scssFilepath, content);

        const jsFilepath = path.resolve(this._publicPath + output.replace(/\.[^/.]+$/, '') + '.scss.js');
        fs.mkdirSync(path.dirname(jsFilepath), {recursive: true});
        fs.writeFileSync(
            jsFilepath,
            'import \'./' + path.basename(scssFilepath) + '\';\n'
        );
        return jsFilepath;
    }
}

module.exports = SCSSEntryPointFileWriter;
