<?php

namespace Oro\Bundle\WindowsBundle\Tests\Unit\Manager;

use Oro\Bundle\WindowsBundle\Manager\WindowsStateRequestManager;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

class WindowsStateRequestManagerTest extends \PHPUnit\Framework\TestCase
{
    /** @var RequestStack|\PHPUnit\Framework\MockObject\MockObject */
    private $requestStack;

    /** @var WindowsStateRequestManager */
    private $manager;

    protected function setUp(): void
    {
        $this->requestStack = $this->createMock(RequestStack::class);

        $this->manager = new WindowsStateRequestManager($this->requestStack);
    }

    public function testGetDataMissingRequest()
    {
        $this->expectException(\InvalidArgumentException::class);
        $this->expectExceptionMessage('Missing $request');

        $this->requestStack->expects($this->once())
            ->method('getCurrentRequest')
            ->willReturn(null);

        $this->manager->getData();
    }

    public function testGetDataMissingData()
    {
        $this->expectException(\InvalidArgumentException::class);
        $this->expectExceptionMessage('Missing data in $request');

        $request = new Request();
        $this->requestStack->expects($this->once())
            ->method('getCurrentRequest')
            ->willReturn($request);

        $this->manager->getData();
    }

    public function testGetDataMissingUrl()
    {
        $this->expectException(\InvalidArgumentException::class);
        $this->expectExceptionMessage('Missing url in $data');

        $request = new Request([], ['data' => []]);
        $this->requestStack->expects($this->once())
            ->method('getCurrentRequest')
            ->willReturn($request);

        $this->manager->getData();
    }

    public function testCleanUrlIsEmpty()
    {
        $this->expectException(\InvalidArgumentException::class);
        $this->expectExceptionMessage('$cleanUrl is empty');

        $request = new Request([], ['data' => ['url' => 'localhost']], [], [], [], ['SCRIPT_NAME' => 'localhost']);
        $this->requestStack->expects($this->once())
            ->method('getCurrentRequest')
            ->willReturn($request);

        $this->manager->getData();
    }

    public function testGetData()
    {
        $request = new Request([], ['data' => ['url' => 'localhost/path']], [], [], [], ['SCRIPT_NAME' => 'localhost']);
        $this->requestStack->expects($this->once())
            ->method('getCurrentRequest')
            ->willReturn($request);

        $this->assertEquals(['url' => 'localhost/path', 'cleanUrl' => '/path'], $this->manager->getData());
    }

    /**
     * @dataProvider uriDataProvider
     */
    public function testGetUri(string $expected, array $data)
    {
        $this->assertStringStartsWith($expected, $this->manager->getUri($data));
    }

    public function uriDataProvider(): array
    {
        return [
            'simple path' => ['/path', ['cleanUrl' => '/path']],
            'autogenerated wid' => ['/path?_widgetContainer=form&_wid=', ['cleanUrl' => '/path', 'type' => 'form']],
            'predefined wid' => [
                '/path?_widgetContainer=form&_wid=123456',
                ['cleanUrl' => '/path', 'type' => 'form', 'wid' => '123456'],
            ],
            'url contains container' => [
                '/path?_widgetContainer=container',
                ['cleanUrl' => '/path?_widgetContainer=container', 'type' => 'form', 'wid' => '123456'],
            ],
            'query' => [
                '/path?param=value&_widgetContainer=form&_wid=123456',
                ['cleanUrl' => '/path?param=value', 'type' => 'form', 'wid' => '123456'],
            ],
        ];
    }

    public function testDataEmpty()
    {
        $this->expectException(\InvalidArgumentException::class);
        $this->expectExceptionMessage('cleanUrl is missing');

        $this->manager->getUri([]);
    }
}
